/// <reference path="../index/node.d.ts"/>
/// <reference path="../index/doctrine.d.ts"/>
var finder = require('./finder');

function flatten(arrays) {
    var acc = [];

    arrays.forEach(function (a) {
        return a.forEach(function (el) {
            return acc.push(el);
        });
    });

    return acc;
}

function find(values, predicate) {
    for (var i = 0; i < values.length; i++) {
        if (predicate(values[i]))
            return values[i];
    }

    return null;
}

function is_any(t) {
    switch (t.type) {
        case 'UnionType':
        case 'AllLiteral':
        case 'NullableLiteral':
        case 'NullLiteral':
        case 'UndefinedLiteral':
            return true;
        default:
            return false;
    }
}

function simplify(types) {
    var unique = types.filter(function (t) {
        return !is_any(t);
    });
    var anys = types.filter(is_any);

    if (anys.length > 0)
        unique.push(anys[0]);

    return unique;
}

function inline(t) {
    if (!t)
        return null;

    switch (t.type) {
        case 'NameExpression':
            if (finder.file(t.name)) {
                var member = finder.member(t.name);
                if (!member)
                    return t;

                var typedef = find(member.jsdoc.tags, function (t) {
                    return t.title === 'typedef';
                });
                if (!typedef)
                    return t;

                var type = typedef.type;
                if (!type)
                    return t;

                // Only inline union types
                if (type.type === 'UnionType')
                    return exports.inline(type);
                else
                    return t;
            } else
                return t;

        case 'AllLiteral':
        case 'NullableLiteral':
        case 'NullLiteral':
        case 'UndefinedLiteral':
        case 'VoidLiteral':
            return t;

        case 'TypeApplication':
            return {
                type: t.type,
                expression: t.expression,
                applications: t.applications.map(exports.inline)
            };
        case 'OptionalType':
        case 'NullableType':
        case 'NonNullableType':
        case 'RestType':
            return {
                type: t.type,
                expression: exports.inline(t.expression)
            };
        case 'FunctionType':
            return {
                type: t.type,
                params: t.params.map(exports.inline),
                result: exports.inline(t.result)
            };
        case 'RecordType':
            return {
                type: t.type,
                fields: t.fields.map(exports.inline)
            };
        case 'ArrayType':
            return {
                type: t.type,
                elements: t.elements.map(exports.inline)
            };
        case 'FieldType':
            return {
                type: t.type,
                key: t.key,
                value: exports.inline(t.value)
            };
        case 'ParameterType':
            return {
                type: t.type,
                name: t.name,
                expression: exports.inline(t.expression)
            };

        case 'UnionType':
            return {
                type: t.type,
                elements: t.elements.map(exports.inline)
            };
        default:
            throw new Error('Unknown type expression ' + t.type);
    }
}
exports.inline = inline;

/**
* Eliminate union types by splitting into multiple types
* @param t
*/
function unload(t) {
    if (!t)
        return [null];

    switch (t.type) {
        case 'NameExpression':
        case 'AllLiteral':
        case 'NullableLiteral':
        case 'NullLiteral':
        case 'UndefinedLiteral':
        case 'VoidLiteral':
        case 'TypeApplication':
            return [t];

        case 'OptionalType':
        case 'NullableType':
        case 'NonNullableType':
        case 'RestType':
            var acc = [];

            exports.unload(t.expression).forEach(function (e) {
                acc.push({
                    type: t.type,
                    expression: e
                });
            });

            return acc;
        case 'FunctionType':
            var params = exports.outer(t.params.map(exports.unload));
            var result = exports.unload(t.result);
            var acc = [];

            params.forEach(function (p) {
                result.forEach(function (r) {
                    acc.push({
                        type: t.type,
                        params: p,
                        result: r
                    });
                });
            });

            return acc;
        case 'RecordType':
            var fields = exports.outer(t.fields.map(exports.unload));
            var acc = [];

            fields.forEach(function (f) {
                acc.push({
                    type: t.type,
                    fields: f
                });
            });

            return acc;
        case 'ArrayType':
            var elements = exports.outer(t.elements.map(exports.unload));
            var acc = [];

            elements.forEach(function (e) {
                acc.push({
                    type: t.type,
                    elements: e
                });
            });

            return acc;
        case 'FieldType':
            var value = exports.unload(t.value);
            var acc = [];

            value.forEach(function (v) {
                acc.push({
                    type: t.type,
                    key: t.key,
                    value: v
                });
            });

            return acc;
        case 'ParameterType':
            var expression = exports.unload(t.expression);
            var acc = [];

            expression.forEach(function (e) {
                acc.push({
                    type: t.type,
                    name: t.name,
                    expression: e
                });
            });

            return acc;

        case 'UnionType':
            var simple = simplify(t.elements);
            var unions = simple.map(exports.unload);

            return flatten(unions);
        default:
            throw new Error('Unknown type expression ' + t.type);
    }
}
exports.unload = unload;

function outer(arrays) {
    if (arrays.length === 0)
        return [[]];
    else {
        var next = exports.outer(arrays.slice(1));
        var acc = [];

        arrays[0].forEach(function (head) {
            next.forEach(function (tail) {
                acc.push([head].concat(tail));
            });
        });

        return acc;
    }
}
exports.outer = outer;
//# sourceMappingURL=disunion.js.map
